/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODE "F(304)"

/*
 * memtrack.c --
 *
 *      Utility module for tracking memory allocated and/or 
 *      locked by the monitor. 
 *     
 *      
 */

#ifdef linux
/* Must come before any kernel header file --hpreg */
#   include "driver-config.h"

#   include <linux/string.h> /* memset() in the kernel */
#else
#   undef PAGE_SIZE        /* Redefined in ntddk.h, and we use that defn. */
#   undef PAGE_SHIFT
#   include <ntddk.h>
#endif

#include "machine.h"
#include "vm_types.h"
#include "vm_assert.h"
#include "memtrack.h"
#include "vmx86.h"
#include "hostif.h"
#include "vm_atomic.h"


/*
 * MemTrack - We track memory using a two-level page table
 * like structure. We allocate the first level in the granularity 
 * of pages and pack as many MemTrackEntires as possible in a page. 
 * We currently handle 4 gig's worth of pages.  -- edward
 */

#define MAX_TRACE_PAGE_DIR \
   CEILING(1 << (32 - PAGE_SHIFT), PAGE_SIZE / sizeof (MemTrackEntry))

#define HASH_TABLE_SIZE    16384
#define HASH_TABLE_ENTRIES_PER_PAGE (PAGE_SIZE / sizeof (void *))
#define HASH_TABLE_PAGES   (HASH_TABLE_SIZE / HASH_TABLE_ENTRIES_PER_PAGE)


/* Anonymous pages to be recycled are kept in a linked list. To make
 * the list mp safe, the anchor of the list consists of a pointer to
 * the first element and a counter. The counter is incremented each time
 * the list is modified by a process. The pointer and counter are written
 * atomically with a 64-bit operation. The write operation succeeds only
 * if the list has not been modified since the values have been read.
 * This scheme can theoretically fail if a process gets interrupted
 * while modifying the list, and other processes modify the list 2^32 times
 * (so the counter has wrapped and has the same value again) _and_ the
 * same anonymous page happens to be at the head of the list. This is
 * sufficiently unlikely to happen that we accept this implementation
 * as correct.
 */
typedef struct RecListHead {
   MemTrackEntry *head;  // pointer to first element of list
   uint32 generation;    // generation counter for MP serialization
} RecListHead;

typedef struct MemTrack {
   int numEntriesPerPage;    /* Number of entry per page */
   int numPages;             /* Number of pages stored in the tracker */
   RecListHead reclist;      /* List of pages to be recycled */
   char *mapPages[MAX_TRACE_PAGE_DIR];
   MemTrackEntry **vpnHashTablePages[HASH_TABLE_PAGES]; /* VPN to entry hashtable */
   MemTrackEntry **mpnHashTablePages[HASH_TABLE_PAGES]; /* MPN to entry hashtable */
} MemTrack;

static INLINE MemTrackEntry **
HASH_VPN(MemTrackEntry **hashTablePages[HASH_TABLE_PAGES],
	 VPN vpn) 
{
   uint32 hash = vpn % HASH_TABLE_SIZE;
   int page = hash / HASH_TABLE_ENTRIES_PER_PAGE;
   int offset = hash % HASH_TABLE_ENTRIES_PER_PAGE;
   return hashTablePages[page] + offset;
}

#define HASH_MPN(_ht, _mpn) HASH_VPN((_ht), (_mpn))

/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Init --
 *
 *      Allocate and initialize the memory tracker.
 *
 * Results:
 *      Handle used to access the memtracker.
 *
 * Side effects:
 *      memory allocation.
 *
 *----------------------------------------------------------------------
 */

void *
MemTrack_Init(void)
{
   MemTrack *st;
   int i;

   st = HostIF_AllocKernelMem(sizeof *st, FALSE);
   if (st == NULL) { 
      Warning("MemTrack_Init failed\n");
      return NULL;
   }
   
   memset(st, 0, sizeof *st);
   
   for (i = 0; i < HASH_TABLE_PAGES; i++) {
      VA vpnPage = (VA)HostIF_AllocPage();
      VA mpnPage = (VA)HostIF_AllocPage();

      if (vpnPage == 0 || mpnPage == 0) { 
         Warning("MemTrack_Init failed on hashTablePages %d\n",i);
         // LEAK
         return NULL;
      }
      if (vpnPage & (PAGE_SIZE-1)) {
         Warning("MemTrack_Init unaligned 0x%08x\n", (uintptr_t)st->vpnHashTablePages[i]);
         // LEAK
         return NULL;
      }
      if (mpnPage & (PAGE_SIZE-1)) {
         Warning("MemTrack_Init unaligned 0x%08x\n", (uintptr_t)st->mpnHashTablePages[i]);
         // LEAK
         return NULL;
      }
      st->vpnHashTablePages[i] = (MemTrackEntry **)vpnPage;
      memset((char*)vpnPage, 0, PAGE_SIZE);

      st->mpnHashTablePages[i] = (MemTrackEntry **)mpnPage;
      memset((char*)mpnPage, 0, PAGE_SIZE);
   }
   
   ASSERT(sizeof (MemTrackEntry) <= PAGE_SIZE);
   st->numEntriesPerPage = PAGE_SIZE / sizeof (MemTrackEntry);
   st->numPages = 0;
   memset(st->mapPages, 0, sizeof st->mapPages);

   return (void *)st;
}


/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Add --
 *
 *      Add the specified address to the memory tracker. 
 *
 * Results:
 *      A pointer to the allocated element, or NULL on error.
 *
 * Side effects:
 *      memory allocation.
 *
 *----------------------------------------------------------------------
 */

MemTrackEntry *
MemTrack_Add(void *s,  // memtracker handler
             VPN vpn,  // VPN of entry
             MPN mpn)  // MPN of entry
{
  MemTrack *st = (MemTrack *) s;
  int ind = st->numPages;
  int dirPage = ind / st->numEntriesPerPage;
  int offset = ind % st->numEntriesPerPage;
  MemTrackEntry *ptr;
  MemTrackEntry **head;
  /*
   * Allocate a directory page if needed.
   */
  if (dirPage >= MAX_TRACE_PAGE_DIR) return NULL;

  if (st->mapPages[dirPage] == NULL) {
     st->mapPages[dirPage] = HostIF_AllocPage();
     if (st->mapPages[dirPage] == NULL) return NULL;
  }

  /*
   * Store the page in the tracker 
   */
  ptr = (MemTrackEntry *) st->mapPages[dirPage] + offset;
  ptr->vpn = vpn;  
  ptr->mpn = mpn;  
  ptr->reclist = NULL;
#ifdef _WIN32
  ptr->mdl = NULL;
#endif

  /* Add entry in the VPN hash table. */
  head = HASH_VPN(st->vpnHashTablePages,vpn);
  ptr->vpnHashChain = *head;
  *head = ptr;

  /* Add entry in the MPN hash table. */
  head = HASH_MPN(st->mpnHashTablePages,mpn);
  ptr->mpnHashChain = *head;
  *head = ptr;

  st->numPages++;
  
  return ptr;
}

/*
 *----------------------------------------------------------------------
 *
 * MemTrack_LookupVPN --
 *
 *      Lookup the specified address in the memory tracker. 
 *
 * Results:
 *      A pointer to the allocated element, or NULL if not there.
 *
 *----------------------------------------------------------------------
 */

MemTrackEntry *
MemTrack_LookupVPN(void *s,  // memtracker handler
		   VPN vpn)  // Value to lookup
{
  MemTrack *st = (MemTrack *) s;
  MemTrackEntry *nextPtr;
    
  nextPtr = *HASH_VPN(st->vpnHashTablePages,vpn);
  while (nextPtr != NULL) {
    if (nextPtr->vpn == vpn) {
      return nextPtr;
    }
    nextPtr = nextPtr->vpnHashChain;
  }
  return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * MemTrack_LookupMPN --
 *
 *      Find a page given its mpn in the memory tracker. 
 *
 * Results:
 *      A pointer to the allocated element, or NULL if not there.
 *
 *----------------------------------------------------------------------
 */
MemTrackEntry *
MemTrack_LookupMPN(void *s, MPN mpn)
{
  MemTrack *st = (MemTrack *) s;
  MemTrackEntry *nextPtr;
    
  nextPtr = *HASH_MPN(st->mpnHashTablePages,mpn);
  while (nextPtr != NULL) {
    if (nextPtr->mpn == mpn) {
      return nextPtr;
    }
    nextPtr = nextPtr->mpnHashChain;
  }
  return NULL;
}


/*
 *----------------------------------------------------------------------
 *
 * MemTrack_RecyclePage --
 *
 *      Mark page of this memtracker entry as unused so it can be
 *      reused. The page must not be locked.
 *      Atomically inserts e into the list of recycled pages.
 *
 * Results:
 *      none
 *
 *----------------------------------------------------------------------
 */
void
MemTrack_RecyclePage(void *s, MemTrackEntry *e)
{
   MemTrack *tracker = (MemTrack*)s;
   RecListHead oldhead, newhead;
   Bool success;

   ASSERT(e->mpn == 0);
   // Atomically add e to the list of recycled pages.
   do {
      oldhead = tracker->reclist;
      e->reclist = oldhead.head;
      newhead.head = e;
      newhead.generation = oldhead.generation + 1;
      success = Atomic_CMPXCHG64((Atomic_uint64 *)&tracker->reclist,
                   (S_uint64*)&oldhead, (S_uint64*)&newhead);
   } while (!success);
}


/*
 *----------------------------------------------------------------------
 *
 * MemTrack_GetRecycledPage --
 *
 *      Find an unused page that can be recycled. Atomically remove
 *      the first entry in the list of recycled pages.
 *
 * Results:
 *      Pointer to memtracker entry of unused page, or null if
 *      there are no unused pages.
 *
 *----------------------------------------------------------------------
 */
MemTrackEntry *
MemTrack_GetRecycledPage(void *s)
{
   MemTrack *tracker = (MemTrack *)s;
   RecListHead oldhead, newhead;
   Bool success;

   oldhead = tracker->reclist;
   if (oldhead.head != NULL) {
      do {
	 newhead.head = oldhead.head->reclist;
	 newhead.generation = oldhead.generation + 1;
	 success = Atomic_CMPXCHG64((Atomic_uint64*)&tracker->reclist,
		      (S_uint64*)&oldhead, (S_uint64 *)&newhead);
      } while (!success);
      oldhead.head->reclist = NULL; // not necessary, but safer
   }
   return oldhead.head;
}



/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Scan --
 *
 *      Scan the stored pages using a provide search function. 
 *
 * Results:
 *      The return result of the searchFunc
 *
 * Side effects:
 *      None
 *
 *----------------------------------------------------------------------
 */

void *
MemTrack_Scan(void *s,  // Handle for memtracker
        void *arg,  // Argument to searchFunc
        void *(*searchFunc)(void *arg, MemTrackEntry *ptr))
{
  MemTrack *st = (MemTrack *) s;
  int ind;
  int dirPage;
  MemTrackEntry *ptr;
  void *ret;

  /* 
   * Call searchFunc with every address stored in the tracker
   * or until it returns a non-zero value.
   */
  for (ind = 0; ind < st->numPages; ind++) {
    dirPage = ind / st->numEntriesPerPage;
    ptr = (MemTrackEntry *) st->mapPages[dirPage] + 
          (ind % st->numEntriesPerPage);
    ret = searchFunc(arg, ptr);
    if (ret) return ret;
  }
  return NULL;
}




/*
 *----------------------------------------------------------------------
 *
 * MemTrack_Cleanup --
 *
 *      Cleanup all resources allocated for the memtracker. For
 *      all pages in the tracker call the user provided free function.
 *
 * Results:
 *      Number of pages in the tracker. 
 *
 * Side effects:
 *      Memory free
 *
 *----------------------------------------------------------------------
 */

int 
MemTrack_Cleanup(void *s,   // Mem tracker handle
                 void (*CleanUp)(void *clientData,MemTrackEntry*),
                 void *clientData)   // free functions.
{
  MemTrack *st = (MemTrack *) s;
  int ind;
  int dirPage;
  MemTrackEntry *ptr;
  int count = 0;
  int i;

  for (ind = 0; ind < st->numPages; ind++) {
    dirPage = ind / st->numEntriesPerPage;
    ptr = (MemTrackEntry *) st->mapPages[dirPage] + 
          (ind % st->numEntriesPerPage);
    CleanUp(clientData,ptr);
    count++;
  }
  /*
   * Free any directory page number and then free the tracker
   * structure itself 
   */
  for (dirPage = 0; dirPage < MAX_TRACE_PAGE_DIR; dirPage++) {
     if (st->mapPages[dirPage]) { 
        HostIF_FreePage(st->mapPages[dirPage]);
        st->mapPages[dirPage] = NULL;
     }
  }
  
  for (i=0;i<HASH_TABLE_PAGES;i++) { 
     if (st->vpnHashTablePages[i]) {
        HostIF_FreePage(st->vpnHashTablePages[i]);
        st->vpnHashTablePages[i] = NULL;
     }
     if (st->mpnHashTablePages[i]) {
        HostIF_FreePage(st->mpnHashTablePages[i]);
        st->mpnHashTablePages[i] = NULL;
     }
  }

  HostIF_FreeKernelMem(st);
  return count;
}

