/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODE "F(302)"


/*
 * task.c --
 *
 *      Task initialization and switching routines between the host
 *      and the monitor. 
 *     
 *      A task switch: 
 *          -saves the EFLAGS,CR0,CR2,CR4, and IDT
 *          -jumps to code on the shared page
 *              which saves the registers, GDT and CR3
 *              which then restores the registers, GDT and CR3
 *          -restores the IDT,CR0,CR2,CR4 and EFLAGS
 *                
 *      This file is pretty much independent of the host OS.
 *      
 */

#ifdef linux
/* Must come before any kernel header file --hpreg */
#   include "driver-config.h"

#   ifdef USE_PERFCTRS_HOSTED
#      include "perfctr.h"
#   endif

#   ifdef KERNEL_2_1
#      define EXPORT_SYMTAB
#   endif
#endif

#include "vm_types.h"
#include "vm_assert.h"
#include "modulecall.h"
#include "vm_assert.h"
#include "vmx86.h"
#include "task.h"
#include "vm_asm.h"
#include "vtrace.h"
#include "cpuid.h"
#include "hostif.h"
/* On Linux, must come before any inclusion of asm/page.h --hpreg */
#include "hostKernel.h"


/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

static Bool hostUsesSysenter = FALSE;
int 
Task_InitCrosspage(VMDriver *vm,          // IN
                   InitBlock *initParams) // IN: Initial params from the VM 
{
   LA baseLinearAddr = HOST_KERNEL_VA_2_LA(0);
   uint32 cr3reg;
   Vcpuid vcpuid;

   ASSERT(sizeof (VMCrossPage) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
      /*
       * The physical memory associated to the crosspage has already been
       * locked in MonitorAllocateMemory() --hpreg
       */
      VMCrossPage *p = HostIF_UserToDriverPtr(vm, initParams->crosspage[vcpuid]);
      if (p == NULL) {
         return 1;
      }
      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
      p->hostContext.task.cr3 = cr3reg;
      p->hostVA = p;
      p->hostContextVA = (uintptr_t)&p->hostContext;
      p->crosspageLinearAddr = baseLinearAddr + (uintptr_t)p;
      {
         DTR idtrReg;
         GET_IDT(idtrReg);
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
      }

      {
         uint16 trReg;
         GET_TR(trReg);
         p->hostContext.tr = trReg;
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;
   }
   return 0;
}


static Bool
DisableNMIDelivery(volatile uint32 *regPtr) // IN/OUT
{
   uint32 reg;
   
   reg = *regPtr;
   if ((APIC_LVT_DELVMODE(reg) == APIC_LVT_DELVMODE_NMI) &&
       (! APIC_LVT_ISMASKED(reg))) {
      *regPtr = reg | APIC_LVT_MASK;
      return TRUE;
   }
   return FALSE;
}


static void
DisableNMI(VMDriver *vm,     // IN
           Bool *lint0NMI,   // OUT
           Bool *lint1NMI,   // OUT
           Bool *pcNMI,      // OUT
           Bool *thermalNMI) // OUT 
{
   if (vm->hostAPIC) {
      *lint0NMI = DisableNMIDelivery(&APIC_LINT0_REG(vm->hostAPIC));
      *lint1NMI = DisableNMIDelivery(&APIC_LINT1_REG(vm->hostAPIC));
      *pcNMI = DisableNMIDelivery(&APIC_PC_REG(vm->hostAPIC));

      /*
       * The LVT thermal monitor register was introduced
       * in Pentium 4 and Xeon processors.
       */

      if (APIC_MAX_LVT(vm->hostAPIC) >= 5) {
         *thermalNMI = DisableNMIDelivery(&APIC_THERM_REG(vm->hostAPIC));
      } else {
         *thermalNMI = FALSE;
      }
   }
}


static void
RestoreNMI(VMDriver *vm,    // IN
           Bool lint0NMI,   // IN
           Bool lint1NMI,   // IN
           Bool pcNMI,      // IN
           Bool thermalNMI) // IN
{
#define RestoreNMIDelivery(cond, apicr)				\
   do {								\
      if (cond) {						\
         uint32 reg;						\
	 							\
         reg = apicr;						\
         apicr = reg & ~APIC_LVT_MASK;				\
      }								\
   } while (0)

   RestoreNMIDelivery(lint0NMI, APIC_LINT0_REG(vm->hostAPIC));
   RestoreNMIDelivery(lint1NMI, APIC_LINT1_REG(vm->hostAPIC));
   RestoreNMIDelivery(pcNMI, APIC_PC_REG(vm->hostAPIC));
   RestoreNMIDelivery(thermalNMI, APIC_THERM_REG(vm->hostAPIC));
#undef RestoreNMIDelivery
}


typedef struct SysenterState {
   Selector	cs;
   uint32	esp;
   uint32	eip;
} SysenterState;


/*
 * Sysenter state handling. Writing MSRs is a costly operation, so we
 * try to conserve on unnecessary writes.
 *
 * We assume that if the host has a 0 in its sysenter CS msr, it is
 * completely sysenter unaware. Therefore, we can let the monitor trample
 * on other MSRs as it pleases.
 */
static INLINE_SINGLE_CALLER void
SaveSysenterState(SysenterState *st,      // IN
                  VMCrossPage *crosspage) // IN
{
   if (hostUsesSysenter) {
      st->cs = (Selector) HostIF_RDMSR(MSR_SYSENTER_CS);
      ASSERT(st->cs != 0);
      if (crosspage->monUsesSysenter) {
	 st->esp = (uint32) HostIF_RDMSR(MSR_SYSENTER_ESP);
	 st->eip = (uint32) HostIF_RDMSR(MSR_SYSENTER_EIP);
      } else {
	 HostIF_WRMSR(MSR_SYSENTER_CS, 0);
      }
   }
}


static INLINE_SINGLE_CALLER void
RestoreSysenterState(SysenterState *st,      // IN
                     VMCrossPage* crosspage) // IN
{
   if (hostUsesSysenter) {
      ASSERT(st->cs != 0);
      HostIF_WRMSR(MSR_SYSENTER_CS, st->cs);
      if (crosspage->monUsesSysenter) {
	 /* The monitor smashed these registers too. */
	 HostIF_WRMSR(MSR_SYSENTER_ESP, st->esp);
	 HostIF_WRMSR(MSR_SYSENTER_EIP, st->eip);
      }
   } else {
      if (crosspage->monUsesSysenter) {
	 /*
	  * The monitor smashed multiple registers, but since the host
	  * appears sysenter unaware, we'll do the minimal amount of work
	  * here to ensure that host sysenters #GP.
	  */
	 HostIF_WRMSR(MSR_SYSENTER_CS, 0);
      }
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * SwitchToMonitor --
 *
 *      Wrapper that calls code to switch from the host to the monitor.
 *
 *      The wrapper ends with a long jump after having
 *      pushed EAX and EBX on the stack. Compensate for these
 *      two registers, and then for the retval and argument on
 *      on the stack
 *
 *      Note the dummy 4th argument on the stack compensates
 *      for the absence of segment information pushed on a
 *      near jump.
 *
 * Side effects:
 *      None for the module
 *
 * Bugs: EAX is caller-saved, no need to save/restore
 *       should use caller-saved EDX rather than EBX as temp
 *       in Task_Switch()
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SwitchToMonitor(char *codePtr,
                ContextInfo *src,
                ContextInfo *dst,
                uint32 newDstVA)
{
#ifdef __GNUC__
   uint32 dummy1;
   uint32 dummy2;

   /* Checked against the Intel manual and GCC --hpreg */
   __asm__ __volatile__(
      "pushl %2"       "\n\t"
      "pushl %3"       "\n\t"
      "pushl %4"       "\n\t"
      "pushl %5"       "\n\t"
      "call *%6"       "\n\t"
      "popl %0"        "\n\t"
      "popl %1"        "\n\t"
      "addl %7, %%esp"
      : "=&b" (dummy1),
        "=&a" (dummy2)
      : "g" (newDstVA),
        "g" (dst),
        "g" (src),
        "g" (0),
        "rm" (codePtr),
        "g" ((1 + 1 + 3) * 4) // retval + segment + 3 arguments pushed
      : "memory", "cc"
   );
#elif _MSC_VER
   __asm push newDstVA
   __asm push dst
   __asm push src
   __asm push 0
   __asm call codePtr
   __asm pop ebx
   __asm pop eax
   __asm add esp, 20
#else
#   error No compiler defined for SwitchToMonitor
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch(VMDriver *vm,  // IN
            Vcpuid vcpuid) // IN
{
   uint32      monContextVA;     
   DTR         gdtreg;
   Descriptor *desc;
   uint32      flags,cr0reg, cr2reg,cr4reg;
   uint32      dr0reg, dr1reg, dr2reg, dr3reg, dr6reg, dr7reg;
   SysenterState sysenterState;
   Selector    gs, fs;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   Bool lint0NMI = FALSE;
   Bool lint1NMI = FALSE;
   Bool pcNMI = FALSE;
   Bool thermalNMI = FALSE;
   VMCrossPage *crosspage = vm->crosspage[vcpuid];
  
   /*
    * monContextVA points to the same location as
    * vm->crosspage->monContext, only refers to it
    * in the monitor's virtual address space
    */
   monContextVA = (uint32)
      (VPN_2_VA(MODULECALL_CROSS_PAGE_START)+ offsetof(VMCrossPage,monContext)); 
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   DisableNMI(vm, &lint0NMI, &lint1NMI, &pcNMI, &thermalNMI);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();


   VTRACE_SET(0,0x41000001);
   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);

   /*
    * We don't know what the host has been doing with the
    * debug registers, so we save them here and
    * conditionally restore them when we come back.
    *
    * We could do better by saving and restoring the
    * registers on the monitor side only when the monitor
    * needs to use them, but that will be a different
    * change.
    */

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(dr##n##reg); \
                crosspage->shadowDR[n] = dr##n##reg;

   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);

#undef SAVE_DR_SHADOW

   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR0
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(dr7reg & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

   GET_GDT(gdtreg); 
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
   gs = GET_GS();
   fs = GET_FS();

#ifdef KERNEL_2_1
   { 
     /*
      * Be sure that the CR4_PGE (global bit enable) is cleared.
      */
     uint32 new_cr4 = cr4reg & ~CR4_PGE;
     SET_CR4(new_cr4);
   }
#endif
#ifdef WINNT_DDK
   { 
     /*
      * Clear cr4 so only the 4MB pages are supported. 
      */
     uint32 new_cr4 = CR4_PSE;

      /*
       * We need to preserve the PAE attribute, otherwise we will reset
       * the host immediately.
       */
      if (cr4reg & CR4_PAE) {
         new_cr4 |= CR4_PAE; 
      }

     SET_CR4(new_cr4);
   }
#endif
   SaveSysenterState(&sysenterState, crosspage);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
   crosspage->hostContext.tr = trReg;
   if (trReg) {
     vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */

   *(uint16*) (&crosspage->hostContext.jump[4]) = GET_CS();
   crosspage->hostContext.task.cs = GET_CS();
   
   /*
    * See comment for SwitchToMonitor()
    */

   SwitchToMonitor(&crosspage->contextSwitchCode[0],
                  &crosspage->hostContext,
                  &crosspage->monContext,
                  monContextVA);

   RestoreSysenterState(&sysenterState, crosspage);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);  
   
   SET_FS(fs);
   SET_GS(gs);

#define RESTORE_DR(n) if (dr##n##reg != crosspage->shadowDR[n]) \
                              SET_DR##n(dr##n##reg)
   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR

   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_INTR) {
      VTRACE_SET(0,0x42000001|((crosspage->args[0]&0xff)<<4));
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 VTRACE_SET(0,0x4f000001);
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }

      VTRACE_SET(0,0x43000001);

   } else {
     VTRACE_SET(0,0x44000001);
   }
   
   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint0NMI, lint1NMI, pcNMI, thermalNMI);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}


/*
 *----------------------------------------------------------------------
 * Task_ProbeSysenterMSR --
 *
 *    This function computes sysenterMSR. Note that we assume that if a
 *    host is ever going to use the sysenter MSRs, they will have been
 *    initialized here; if some future host lazily sets up the sysenter
 *    MSRs, this code will cause trouble.
 *
 *----------------------------------------------------------------------
 */
void
Task_ProbeSysenterMSR(void)
{
   CPUOem cpuOem = CPUID_GetOEM();
   uint32 features = CPUID_GetFeatures();
   uint32 version = CPUID_GetVersion();

   if (cpuOem == CPUID_OEM_INTEL && CPUID_FAMILY(version) == 6 &&
       CPUID_MODEL(version) < 3 && CPUID_STEPPING(version) < 3) {

      /* The Intel manual warns that the PPro erroneously sets the SEP
       * CPUID bit.
       */
      Warning("VMMON: CPU too old to use sysenter (ok).\n");
   } else if (features & CPUID_FEATURE_COMMON_SEP) {
      uint64 tmpMSR;
      /* The CPU appears to support sysenter. Does the host actually
       * use it?
       */
      tmpMSR = HostIF_RDMSR(MSR_SYSENTER_CS);
      hostUsesSysenter = (tmpMSR != 0);
   }
}
