/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 *  vmx86.h - Platform independent data and interface for supporting 
 *            the vmx86 device driver. 
 */

#ifndef VMX86_H
#define VMX86_H

#define INCLUDE_ALLOW_MODULE
#include "includeCheck.h"


#include "machine.h"
#include "modulecall.h"
#include "vm_types.h"
#include "initblock.h"
#include "memmgmt.h"
#include "wslimits.h"


/*
 * VMDriver - the main data structure for the driver side of a 
 *            virtual machine.
 */ 

#define NUM_PASSTHROUGH_IRQS 10

#ifdef WINNT_DDK
struct LowMemMap;
#endif

typedef struct VMDriver {
   int                 count;        /* reference count */
   int                 id;           /* unique VM ID */
   struct VMDriver    *nextDriver;   /* Next on list of all VMDrivers */
   uintptr_t           processID;    /* User level process handle. */
   void               *uniqHandle;   /* Unique handle of driver */

   int                 logFD;        /* File descriptor for log messages */
   Vcpuid              numVCPUs;     /* Number of vcpus in VM. */
   struct VMHost      *vmhost;       /* Host-specific fields. */

   /* Pointers to the crossover pages shared with the monitor. */
   struct VMCrossPage *crosspage[MAX_INITBLOCK_CPUS];
   struct page        *crosspagePage;
#ifdef WINNT_DDK
   void               *crossPageMDL;
   struct LowMemMap   *lowMemMaps;
#endif
   volatile uint32   (*hostAPIC)[4]; /* kseg pointer to host APIC */

   void               *memtracker;   /* Memory tracker pointer */
   struct PhysTracker *physTracker;
   int                 releaseCount;
   Bool                checkFuncFailed;
   struct PerfCounter *perfCounter;
   VMX86Stats	       stats;
   VMMemMgmtInfo       memMgmtInfo;
   WslimitsInfo        wslimitsInfo;
} VMDriver;


extern VMDriver *Vmx86_CreateVM(void *uniqHandle, uintptr_t processId, 
				WslimitsInfo *minfo);
extern VMDriver *Vmx86_BindVM(int id);
extern int Vmx86_ReleaseVM(VMDriver *vm);
extern int Vmx86_InitVM(VMDriver *vm, InitBlock *initParams);
extern int Vmx86_LateInitVM(VMDriver *vm);
extern int Vmx86_RunVM(VMDriver *vm, Vcpuid vcpuid);
extern void Vmx86_SetStartTime(VmTimeStart *st);
extern uint32 Vmx86_GetkHzEstimate(VmTimeStart *st);
extern VMDriver *Vmx86_GetVMforProcess(uintptr_t processId);
extern MPN Vmx86_LockPage(VMDriver *vm, void *addr, Bool userPage);
extern int Vmx86_UnlockPage(VMDriver *vm, void *addr, Bool userPage);
extern MPN Vmx86_GetRecycledPage(VMDriver *vm);
extern int Vmx86_ReleaseAnonPage(VMDriver *vm, MPN mpn);

extern int32 Vmx86_GetNumVMs(void);
extern int32 Vmx86_GetTotalMemUsage(void);
extern int32 Vmx86_GetLockedPagesLimit(void);
extern Bool Vmx86_SetLockedPagesLimit(uintptr_t limit);
extern void Vmx86_SetLockedPageLimitMonitorStatus(Bool status);
extern uintptr_t Vmx86_ChangeLockedPagesLimit(uintptr_t limit);
extern Bool Vmx86_GetMemInfo(VMDriver *curVM,
                             int32 curVMOnly,
                             VMGetMemInfoArgs *args,
                             int outArgsLength);
extern Bool Vmx86_GetMemInfoCopy(VMDriver *curVM,
				 VMGetMemInfoArgs *args);
extern void Vmx86_SetMemInfo(VMDriver *curVM, VMSetMemInfoArgs *args);
extern void Vmx86_Add2MonPageTable(VMDriver *vm, VPN vpn, MPN mpn,
				   Bool readOnly);
extern Bool Vmx86_PAEEnabled(void);
extern void Vmx86_AckUserCall(VMDriver *vm, Vcpuid vcpuid);
extern void Vmx86_CompleteUserCall(VMDriver *vm, Vcpuid vcpuid);

#ifdef linux
#include <linux/errno.h>
#define USERCALL_RESTART (-ERESTARTNOINTR)
#else
#include <usercall.h>
#define USERCALL_RESTART (USERCALL_NOP)
#endif

#endif 
