/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 * vm_atomic.h --
 *
 *	Atomic power
 */

#ifndef _ATOMIC_H_
#define _ATOMIC_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_MONITOR
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"


#include "vm_basic_types.h"


/*
 * There's only one atomic type.
 * We don't have templates.
 */

typedef struct {
   volatile uint32 value;
} Atomic_uint32;

/*
 * 64 bit atomic type
 */
typedef struct {
   volatile uint32 lowValue;
   volatile uint32 highValue;
} Atomic_uint64;

/*
 * non-atomic type that matches Atomic_uint64
 */
typedef struct {
   uint32 lowValue;
   uint32 highValue;
} S_uint64;

/*
 * Atomic object with a modification counter(MC). Use an atomic uint64 
 * variable to represent a uint32 concurrent object, with higher 32 bits 
 * used as the modification counter.
 * 
 * The difference between AtomicCon_* and Atomic_* is that it avoids ABA 
 * problem, i.e. if a process reads a value A in a shared location, computes
 * a new value, and then attempts a "cmpxchg" operation, the "cmpxchg" may 
 * succeed when it should not, if between the read and the "cmpxchg" some 
 * other process changes the value from A to B and then back to A again.
 * The following structure assures that ABA problem would unlikely to happen.
 */

typedef struct {
   uint32 counter;
   uint32 value;
} MC_uint32;

typedef struct {
   volatile uint32 counter;
   volatile uint32 value;
} AtomicMC_uint32;

/* Convert a volatile int to Atomic_uint32. */
static INLINE Atomic_uint32 *
Atomic_VolatileToAtomic(volatile uint32 *var)
{
   return (Atomic_uint32 *)var;
}

/*
 * All the assembly code is tricky and written conservatively.
 * For example, to make sure gcc won't introduce copies,
 * we force the addressing mode like this:
 *
 *    "xchgl %0, (%1)"
 *    : "=r" (val)
 *    : "r" (&var->value),
 *      "0" (val)
 *    : "memory"
 *
 * - edward
 *
 * Actually - turns out that gcc never generates memory aliases (it
 * still does generate register aliases though), so we can be a bit
 * more agressive with the memory constraints. The code above can be
 * modified like this:
 *
 *    "xchgl %0, %1"
 *    : "=r" (val),
 *      "=m" (var->value),
 *    : "0" (val),
 *      "1" (var->value)
 *
 * The advantages are that gcc can use whatever addressing mode it
 * likes to access the memory value, and that we dont have to use a
 * way-too-generic "memory" clobber as there is now an explicit
 * declaration that var->value is modified.
 *
 * see also /usr/include/asm/atomic.h to convince yourself this is a
 * valid optimization.
 *
 * - walken
 */


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Read --
 *
 *      Read
 *
 * Results:
 *      The value of the atomic variable.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE uint32
Atomic_Read(Atomic_uint32 *var)
{
   return var->value;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Write --
 *
 *      Write
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Write(Atomic_uint32 *var, uint32 val)
{
   var->value = val;
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_ReadWrite --
 *
 *      Read followed by write
 *
 * Results:
 *      The value of the atomic variable before the write.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE uint32
Atomic_ReadWrite(Atomic_uint32 *var, uint32 val)
#ifdef __GNUC__
{
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "xchgl %0, %1"
#   if VM_ASM_PLUS
      : "=r" (val),
	"+m" (var->value)
      : "0" (val)
#   else
      : "=r" (val),
	"=m" (var->value)
      : "0" (val),
	"1" (var->value)
#   endif
   );

   return val;
}
#elif _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4035)		// disable no-return warning
{
   __asm mov eax, val
   __asm mov ebx, var
   __asm xchg [ebx]Atomic_uint32.value, eax
   // eax is the return value, this is documented to work - edward
}
#pragma warning(pop)
#else
#error No compiler defined for Atomic_ReadWrite
#endif


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_ReadIfEqualWrite --
 *
 *      Compare exchange: Read variable, if equal to oldVal, write newVal
 *
 * Results:
 *      The value of the atomic variable before the write.
 *
 * Side effects:
 *      The variable may be modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE uint32
Atomic_ReadIfEqualWrite(Atomic_uint32 *var, uint32 oldVal, uint32 newVal)
#ifdef __GNUC__
{
   uint32 val;

   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; cmpxchgl %2, %1"
#   if VM_ASM_PLUS
      : "=a" (val),
	"+m" (var->value)
      : "r" (newVal),
	"0" (oldVal)
#   else
      : "=a" (val),
	"=m" (var->value)
      : "r" (newVal),
	"0" (oldVal) 
     /* 
      * "1" (var->value): results in inconsistent constraints on gcc 2.7.2.3
      * when compiling enterprise-2.2.17-14-RH7.0-update.
      * The constraint has been commented out for now. We may consider doing
      * this systematically, but we need to be sure it is the right thing to
      * do. However, it is also possible that the offending use of this asm
      * function will be removed in the near future in which case we may 
      * decide to reintroduce the constraint instead. hpreg & agesen.
      */
#   endif
      : "cc"
   );

   return val;
}
#elif _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4035)		// disable no-return warning
{
   __asm mov eax, oldVal
   __asm mov ebx, var
   __asm mov ecx, newVal
   __asm lock cmpxchg [ebx]Atomic_uint32.value, ecx
   // eax is the return value, this is documented to work - edward
}
#pragma warning(pop)
#else
#error No compiler defined for Atomic_ReadIfEqualWrite
#endif


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_And --
 *
 *      And
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_And(Atomic_uint32 *var, uint32 val)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; andl %1, %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      : "ri" (val)
#   else
      : "=m" (var->value)
      : "ri" (val),
        "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov eax, val
   __asm mov ebx, var
   __asm lock and [ebx]Atomic_uint32.value, eax
#else
#error No compiler defined for Atomic_And
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Or --
 *
 *      Or
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Or(Atomic_uint32 *var, uint32 val)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; orl %1, %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      : "ri" (val)
#   else
      : "=m" (var->value)
      : "ri" (val),
        "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov eax, val
   __asm mov ebx, var
   __asm lock or [ebx]Atomic_uint32.value, eax
#else
#error No compiler defined for Atomic_Or
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Xor --
 *
 *      Xor
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Xor(Atomic_uint32 *var, uint32 val)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; xorl %1, %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      : "ri" (val)
#   else
      : "=m" (var->value)
      : "ri" (val),
        "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov eax, val
   __asm mov ebx, var
   __asm lock xor [ebx]Atomic_uint32.value, eax
#else
#error No compiler defined for Atomic_Xor
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Add --
 *
 *      Addition
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Add(Atomic_uint32 *var, uint32 val)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; addl %1, %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      : "ri" (val)
#   else
      : "=m" (var->value)
      : "ri" (val),
        "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov eax, val
   __asm mov ebx, var
   __asm lock add [ebx]Atomic_uint32.value, eax
#else
#error No compiler defined for Atomic_Add
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Sub --
 *
 *      Subtraction
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Sub(Atomic_uint32 *var, uint32 val)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; subl %1, %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      : "ri" (val)
#   else
      : "=m" (var->value)
      : "ri" (val),
        "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov eax, val
   __asm mov ebx, var
   __asm lock sub [ebx]Atomic_uint32.value, eax
#else
#error No compiler defined for Atomic_Sub
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Inc --
 *
 *      Increment
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Inc(Atomic_uint32 *var)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; incl %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      :
#   else
      : "=m" (var->value)
      : "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov ebx, var
   __asm lock inc [ebx]Atomic_uint32.value
#else
#error No compiler defined for Atomic_Inc
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Dec --
 *
 *      Decrement
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE void
Atomic_Dec(Atomic_uint32 *var)
{
#ifdef __GNUC__
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; decl %0"
#   if VM_ASM_PLUS
      : "+m" (var->value)
      :
#   else
      : "=m" (var->value)
      : "0" (var->value)
#   endif
      : "cc"
   );
#elif _MSC_VER
   __asm mov ebx, var
   __asm lock dec [ebx]Atomic_uint32.value
#else
#error No compiler defined for Atomic_Dec
#endif
}

/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_FetchAndOr --
 *
 *  Return the old value, atomically bitwise or'ing in a given value.
 *
 *  Note that the technique below can be used to implement FetchAndX(), where
 *  X is an arbitrary binary operation.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE uint32
Atomic_FetchAndOr(Atomic_uint32 *var, int32 val)
{
   uint32 res;
   do {
      res = var->value;
   } while (res != Atomic_ReadIfEqualWrite(var, res, res | val));
   return res;
}

/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_FetchAndAdd, Atomic_FetchAndInc, Atomic_FetchAndDec --
 *    Adjust the atomic variable by a specified amount: val, +1, or -1.
 *    Return the value of the atomic variable prior to the adjustment.
 *
 *    If you have to implement FetchAndAdd() on an architecture other
 *    than x86, this code snippet may be useful:
 *
 *       uint32 res;
 *       do {
 *          res = var->value;
 *       } while (res != Atomic_ReadIfEqualWrite(var, res, res + val));
 *       return res;
 * 
 *-----------------------------------------------------------------------------
 */
static INLINE uint32
Atomic_FetchAndAdd(Atomic_uint32 *var, int32 val)
#ifdef __GNUC__
{
   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; xaddl %0, %1"
#   if VM_ASM_PLUS
      : "=r" (val),
	"+m" (var->value)
      : "0" (val)
#   else
      : "=r" (val),
	"=m" (var->value)
      : "1" (var->value),
	"0" (val)
#   endif
      : "cc"
   );

   return val;
}
#elif _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4035)		// disable no-return warning
{
   __asm mov eax, val
   __asm mov ebx, var
   __asm lock xadd [ebx]Atomic_uint32.value, eax
}
#pragma warning(pop)
#else
#error No compiler defined for Atomic_FetchAndAdd
#endif

static INLINE uint32
Atomic_FetchAndInc(Atomic_uint32 *var)
{
   return Atomic_FetchAndAdd(var, 1);
}
 
static INLINE uint32
Atomic_FetchAndDec(Atomic_uint32 *var)
{
   return Atomic_FetchAndAdd(var, -1);
}

/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_CMPXCHG64 --
 *
 *      Compare exchange: Read variable, if equal to oldVal, write newVal
 *
 * Results:
 *      TRUE if equal, FALSE if not equal
 *
 * Side effects:
 *      The variable may be modified.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE Bool
Atomic_CMPXCHG64(Atomic_uint64 *var, // IN/OUT
                 S_uint64* oldVal,    // IN
                 S_uint64* newVal)    // IN
#ifdef __GNUC__
{
   Bool equal;
   int dummy1, dummy2;

   /* Checked against the Intel manual and GCC --walken */
   __asm__ __volatile__(
      "lock; cmpxchg8b %0\n\t"
      "sete %1\n"
#   if VM_ASM_PLUS
      : "+m" (*var),
	"=qm" (equal),
	"=a" (dummy1),
	"=d" (dummy2)
      : "2" (oldVal->lowValue),
        "3" (oldVal->highValue),
        "b" (newVal->lowValue),
        "c" (newVal->highValue)
#   else
      : "=m" (*var),
	"=qm" (equal),
	"=a" (dummy1),
	"=d" (dummy2)
      : /*"0" (*var),*/
	"2" (oldVal->lowValue),
        "3" (oldVal->highValue),
        "b" (newVal->lowValue),
        "c" (newVal->highValue)
#   endif
      : "cc"
   );
   return equal;
} 
#elif _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4035)		// disable no-return warning
{
   __asm mov esi, var
   __asm mov edx, oldVal
   __asm mov ecx, newVal
   __asm mov eax, [edx]Atomic_uint64.lowValue
   __asm mov edx, [edx]Atomic_uint64.highValue
   __asm mov ebx, [ecx]Atomic_uint64.lowValue
   __asm mov ecx, [ecx]Atomic_uint64.highValue
   __asm lock cmpxchg8b [esi]
   __asm sete al
   __asm movzx eax, al
   // eax is the return value, this is documented to work - edward
}
#pragma warning(pop)
#else
#error No compiler defined for Atomic_CMPXCHG64
#endif

/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Read64 --
 *
 *      Read and store.
 *
 * Results:
 *      The value of the atomic variable.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE void
Atomic_Read64(Atomic_uint64 *var, S_uint64 *val)
#ifdef __GNUC__
{
   /* 
    * Since cmpxchg8b will replace the contents of EDX:EAX with the
    * value in memory if there is no match, we need only execute the
    * instruction once in order to atomically read 64 bits from
    * memory.  The only constraint is that ECX:EBX must have the same
    * value as EDX:EAX so that if the comparison succeeds.  We
    * intentionally don't tell gcc that we are using ebx and ecx as we
    * don't modify them and do not care what value they store.
    */
   /* Checked against the Intel manual and GCC --jeffshel */
   __asm__ __volatile__(
      "mov %%ebx, %%eax\n\t"
      "mov %%ecx, %%edx\n\t"
      "lock; cmpxchg8b %2"
      : "=&a" (val->lowValue),
        "=&d" (val->highValue) 
      : "m" (*var)
      : "cc"
   );
}
#elif _MSC_VER
{
   __asm mov ecx, var
   __asm mov ebx, val
   __asm mov edx, ecx
   __asm mov eax, ebx
   __asm lock cmpxchg8b [ecx]
   __asm mov [ebx]S_uint64.lowValue, eax
   __asm mov [ebx]S_uint64.highValue, edx
}
#else
#error No compiler defined for Atomic_Read64
#endif

/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Read64Ret --
 *
 *      Read and return.
 *
 * Results:
 *      The value of the atomic variable.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE uint64
Atomic_Read64Ret(Atomic_uint64 *var)
#ifdef __GNUC__
{
   uint64 value;
   /* 
    * Since cmpxchg8b will replace the contents of EDX:EAX with the
    * value in memory if there is no match, we need only execute the
    * instruction once in order to atomically read 64 bits from
    * memory.  The only constraint is that ECX:EBX must have the same
    * value as EDX:EAX so that if the comparison succeeds.  We
    * intentionally don't tell gcc that we are using ebx and ecx as we
    * don't modify them and do not care what value they store.
    */
   /* Modified from Atomic_Read64 by mann. */
   __asm__ __volatile__(
      "mov %%ebx, %%eax\n\t"
      "mov %%ecx, %%edx\n\t"
      "lock; cmpxchg8b %1"
      : "=&A" (value)
      : "m" (*var)
      : "cc"
   );
   return value;
}
#elif _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4035)		// disable no-return warning
{
   __asm mov ecx, var
   __asm mov edx, ecx
   __asm mov eax, ebx
   __asm lock cmpxchg8b [ecx]
   // edx:eax is the return value; this is documented to work. --mann
}
#pragma warning(pop)
#else
#error No compiler defined for Atomic_Read64
#endif

/*
 *-----------------------------------------------------------------------------
 *
 * Atomic_Write64 --
 *
 *      Write
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      The variable is modified.
 *
 *-----------------------------------------------------------------------------
 */
static INLINE void
Atomic_Write64(Atomic_uint64 *var, S_uint64 *val)
{
   S_uint64 oldVal;
   while(1) {
      oldVal.lowValue = var->lowValue;
      oldVal.highValue = var->highValue;
      if (Atomic_CMPXCHG64(var, &oldVal, val)) {
         return;
      }
   }
}

static INLINE void
MC_Set(MC_uint32 *mcVal, uint32 count, uint32 val)
{
   mcVal->value = val;
   mcVal->counter = count;
}
 
static INLINE void
MC_Init(MC_uint32 *mcVal, uint32 val)
{
   MC_Set(mcVal, 0, val);
}
 
static INLINE void
MC_AssignVal(MC_uint32* newVal, MC_uint32 *oldVal, uint32 val)
{
   MC_Set(newVal, oldVal->counter + 1, val);
}

static INLINE Bool
AtomicMC_CMPXCHG(AtomicMC_uint32 *conVal,
                 MC_uint32* oldVal, MC_uint32* newVal)
{
   return Atomic_CMPXCHG64((Atomic_uint64*)conVal, 
                           (S_uint64*)oldVal, (S_uint64*)newVal);
}

static INLINE void
AtomicMC_Read(AtomicMC_uint32 *var, MC_uint32 *val)
{
   Atomic_Read64((Atomic_uint64*)var, (S_uint64*)val);
}

static INLINE void
AtomicMC_Write(AtomicMC_uint32 *var, MC_uint32 *val)
{
   Atomic_Write64((Atomic_uint64*)var, (S_uint64*)val);
}

 
#endif // ifndef _ATOMIC_H_
