/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 *
 * vm_types.h --
 *
 *    basic data types.-- part of driver-only distribution.
 */


#ifndef _VM_TYPES_H_
#define _VM_TYPES_H_

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_MONITOR
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#define INCLUDE_ALLOW_VMK_MODULE
#define INCLUDE_ALLOW_DISTRIBUTE
#include "includeCheck.h"

#include "vm_basic_types.h"
#include "vm_assert.h"

#ifndef PAGE_SHIFT
#ifdef VM_I386
#define PAGE_SHIFT    12
#else
/* On IA-64, the default page size is 16K. */
#define PAGE_SHIFT    14
#endif
#endif

#ifndef PAGE_SIZE
#define PAGE_SIZE     (1<<PAGE_SHIFT)
#endif

#ifndef PAGE_MASK
#define PAGE_MASK     (PAGE_SIZE - 1)
#endif


#ifndef NULL
#ifdef  __cplusplus
#define NULL    0
#else
#define NULL    ((void *)0)
#endif
#endif

#define VA_2_VPN(_va)  ((_va) >> PAGE_SHIFT)
#define PTR_2_VPN(_ptr) VA_2_VPN((VA)(_ptr))
#define VPN_2_VA(_vpn) ((_vpn) << PAGE_SHIFT)

/*
 * Notice that we don't cast PA_2_PPN's argument to an unsigned type, because
 * we would lose compile-time checks for pointer operands and byte-sized
 * operands. If you use a signed quantity for _pa, ones may be shifted into the
 * high bits of your ppn.
 */
#define PA_2_PPN(_pa)  ((_pa) >> PAGE_SHIFT)
#define PPN_2_PA(_ppn)    ((_ppn) << PAGE_SHIFT)

static INLINE MA  MPN_2_MA(MPN mpn) { return (MA)mpn << PAGE_SHIFT; }
static INLINE MPN MA_2_MPN(MA  ma)  { return (MPN)(ma >> PAGE_SHIFT); }

typedef uint32 LA;
typedef uint32 LPN;
#define LA_2_LPN(_la)     ((_la)>>PAGE_SHIFT)
#define LPN_2_LA(_lpn)    ((_lpn) << PAGE_SHIFT)

#ifndef PAGE_OFFSET
#define PAGE_OFFSET(_addr)  ((_addr)&(PAGE_SIZE-1))
#endif

#ifndef BYTES_2_PAGES
#define BYTES_2_PAGES(_nbytes) ((_nbytes) >> PAGE_SHIFT)
#endif

#ifndef PAGES_2_BYTES
#define PAGES_2_BYTES(_npages) ((_npages) << PAGE_SHIFT)
#endif

#ifndef MBYTES_2_PAGES
#define MBYTES_2_PAGES(_nbytes) ((_nbytes) << (20 - PAGE_SHIFT))
#endif

#ifndef PAGES_2_MBYTES
#define PAGES_2_MBYTES(_npages) ((_npages) >> (20 - PAGE_SHIFT))
#endif

#define LAST_LPN ((1<<(8 * sizeof(LA) - PAGE_SHIFT)) - 1)

/* Valid bits in a LPN. */
#define LPN_MASK LAST_LPN

typedef unsigned MemHandle;

/*
 *----------------------------------------------------------------------
 *
 * IsGoodBA --
 *
 *      Basic checks for validity of a BA
 *
 * Right now this function only checks that the low 20 bits of a BA
 * are all zeroes, which should be enough to detect people who
 * incorrectly tried to use BA as an arithmetic type.
 *
 * Results:
 *      TRUE iif BA seems valid.
 *----------------------------------------------------------------------
 */

static INLINE Bool
IsGoodBA(BA ba)
{
   return (ba & 0xfffff) == 0;
}

static INLINE unsigned
BA_PAGE_OFFSET(BA ba)
{
   ASSERT(IsGoodBA(ba));
   return (uint32)ba >> (32 - PAGE_SHIFT);
}

/*
 *----------------------------------------------------------------------
 *
 * MemOffset_2_BA --
 *
 *      Construct a BA from a MemHandle and a byte offset
 *
 * Results:
 *      constructed BA
 *----------------------------------------------------------------------
 */

static INLINE BA
MemOffset_2_BA(MemHandle slot, uint64 offset)
{
   return ((uint64)slot << 56) | (offset << 20);
}

/*
 *----------------------------------------------------------------------
 *
 * BA_2_MemHandle --
 *
 *      Extract the MemHandle part of a BA
 *
 * Results:
 *      MemHandle
 *----------------------------------------------------------------------
 */

static INLINE MemHandle
BA_2_MemHandle(BA ba)
{
   ASSERT(IsGoodBA(ba));
   return (MemHandle)(ba >> 56);
}

/*
 *----------------------------------------------------------------------
 *
 * BA_2_Offset --
 *
 *      Extract the byte offset part of a BA
 *
 * Results:
 *      byte offset
 *----------------------------------------------------------------------
 */

static INLINE uint64
BA_2_Offset(BA ba)
{
   ASSERT(IsGoodBA(ba));
   return (ba >> 20) & ((((uint64)1)<<36)-1);	/* 36-bit byte offset */
}

/*
 *----------------------------------------------------------------------
 *
 * MemPage_2_BPN --
 *
 *      Construct a BPN from a MemHandle and a PageNum
 *
 * Results:
 *      constructed BPN
 *----------------------------------------------------------------------
 */

static INLINE BPN
MemPage_2_BPN(MemHandle slot, PageNum page)
{
   return (slot << 24) | page;
}

/*
 *----------------------------------------------------------------------
 *
 * BPN_2_MemHandle --
 *
 *      Extract the MemHandle part of a BPN
 *
 * Results:
 *      MemHandle
 *----------------------------------------------------------------------
 */

static INLINE MemHandle
BPN_2_MemHandle(BPN bpn)
{
   return bpn >> 24;
}

/*
 *----------------------------------------------------------------------
 *
 * BPN_2_PageNum --
 *
 *      Extract the PageNum part of a BPN
 *
 * Results:
 *      PageNum
 *----------------------------------------------------------------------
 */

static INLINE uint32
BPN_2_PageNum(BPN bpn)
{
   return bpn & 0xffffff;		/* 24-bit page offset */
}

/*
 *----------------------------------------------------------------------
 *
 * BA_2_BPN --
 *
 *      Extract the BPN part of a BA
 *
 * Results:
 *      BPN
 *----------------------------------------------------------------------
 */

static INLINE BPN
BA_2_BPN(BA ba)
{
   ASSERT(IsGoodBA(ba));
   return (BPN)(ba >> 32);
}

/*
 *----------------------------------------------------------------------
 *
 * BPN_Offset_2_BA --
 *
 *      Construct a BA from a BPN and a byte offset within that page
 *
 * Results:
 *      constructed BA
 *----------------------------------------------------------------------
 */

static INLINE BA
BPN_Offset_2_BA(BPN bpn, uint32 offset)
{
   ASSERT(offset < PAGE_SIZE);
   return ((uint64)bpn << 32) | (offset << 20);
}

/* 4 Mb pages. */
#define VM_LARGE_PAGE_SHIFT  22
#define VM_LARGE_PAGE_SIZE   (1 << VM_LARGE_PAGE_SHIFT)
#define VM_LARGE_PAGE_MASK   (VM_LARGE_PAGE_SIZE - 1)

typedef uint32 VM_PDE;
typedef uint32 VM_PTE;
typedef uint64 VM_PAE_PDE;
typedef uint64 VM_PAE_PTE;
typedef uint64 VM_PDPTE;

typedef int32  Reg;
typedef uint8 *TCA;

typedef uint32 Vcpuid;             /* VCPU number - also a thread id. */
typedef void  *HostThread;         /* Host's thread identifier. */
typedef uint32 VCPUSet;

#define VCPUID_UP         (0)        // Get rid of this in SMP port
#define VCPUID_PIC VCPUID_UP         /* CPUID on which interrupts are routed */
#define VCPUID_INVALID    (~0)

#define BOOT_VCPU_ID		0
#define IS_BOOT_VCPUID(vcpuid)  ((vcpuid) == BOOT_VCPU_ID)
#define IS_BOOT_VCPU()		IS_BOOT_VCPUID(CurVcpuid())

typedef int64 VmTimeType;          /* Time in microseconds */
typedef int64 VmTimeRealClock;     /* Real clock kept in microseconds */
typedef int64 VmTimeVirtualClock;  /* Virtual Clock kept in CPU cycles */

typedef struct VmTimeStart {
   uint64 count;
   uint64 time;
} VmTimeStart;

#define ACCESS_MODE_READ      (1 << 0)
#define ACCESS_MODE_WRITE     (1 << 1)
#define ACCESS_MODE_READWRITE (ACCESS_MODE_READ | ACCESS_MODE_WRITE)
#define ACCESS_MODE_NONE      (0)
#define ACCESS_MODE_MASK      (ACCESS_MODE_READ | ACCESS_MODE_WRITE)

typedef unsigned AccessMode;
typedef uint8    AccessMode8;

/*
 * IRQ
 */

#define NOIRQ 0xFF
/* 
 * 8 'real' PCI IRQs (PIRQA-PIRQD for 440BX + PIRQE-PIRQH for modern chipsets)
 * + 8 'special' (SMI, SCI, NMI, etc) 
 */
#define MAX_VIRQS 32
#define LEGAL_IRQ(_irq) ( ((_irq) == NOIRQ) || ((_irq) < MAX_VIRQS) )

typedef enum {
   PIRQA = 16,
   PIRQB,
   PIRQC,
   PIRQD,

   PIRQE,
   PIRQF,
   PIRQG,
   PIRQH,

   IRQ_SMI,
   IRQ_SCI,
   IRQ_NMI,

   IRQ_LAST
} PIRQ;

#ifndef _WIN32
#define LOWORD(_dw)   ((_dw) & 0xffff)
#endif

#if defined(MONITOR) || defined (VMKERNEL)
#ifdef _WIN32
#  define LOWORD(_dw)   ((_dw) & 0xffff)
#endif
#define HIWORD(_dw)   (((_dw) >> 16) & 0xffff)

#define LOBYTE(_w)    ((_w) & 0xff)
#define HIBYTE(_w)    (((_w) >> 8) & 0xff)

#define LODWORD(_dw)   ((uint32)((_dw) & 0xffffffff))

#ifdef __GNUC__
#  define HIDWORD(_dw)  ((uint32)(((_dw) >> 32) & 0xffffffff))
#elif _WIN32
static INLINE uint32 HIDWORD(uint64 y)
{
   uint32 *ptr = (uint32 *)&y;
   return ptr[1];
}
#else
#  error Unknown platform
#endif

#define MERGE_LOBYTE(_w, _b)    (((_w) & ~0x00ff) | ((_b) & 0x00ff))
#define MERGE_HIBYTE(_w, _b)    (((_w) & ~0xff00) | (((_b) << 8) & 0xff00))

#define DWORD(_hi, _lo)         ((((uint64)(_hi))<<32) | (((uint64)(_lo)) & 0xffffffff))

#endif /* ifdef MONITOR */

/* XXX move to where genexc is defined */
#define EXC_SRC_INTERNAL  1
#define EXC_SRC_IRQ       2
#define EXC_SRC_INT       3

#ifndef offsetof
#define offsetof(type, member) ( (int) & ((type*)0) -> member )
#endif

#ifndef MIN
#define MIN(_a, _b)   (((_a) < (_b)) ? (_a) : (_b))
#endif

#ifndef MAX
#define MAX(_a, _b)   (((_a) > (_b)) ? (_a) : (_b))
#endif

#define ROUNDUP(x,y)		(((x) + (y) - 1) / (y) * (y))
#define ROUNDDOWN(x,y)		((x) / (y) * (y))
#define ROUNDUPBITS(x, bits)	(((uint32) (x) + MASK(bits)) & ~MASK(bits))
#define ROUNDDOWNBITS(x, bits)	((uint32) (x) & ~MASK(bits))
#define CEILING(x, y)		(((x) + (y) - 1) / (y))
#define MASK(n)			((1 << (n)) - 1)	/* make an n-bit mask */

/* Deposit a field _src at _pos bits from the right, with a length of _len, */
/* into the integer _target. */

#define DEPOSIT_BITS(_src,_pos,_len,_target) { \
	unsigned mask = ((1 << _len) - 1); \
	unsigned shiftedmask = ((1 << _len) - 1) << _pos; \
	_target = (_target & ~shiftedmask) | ((_src & mask) << _pos); \
}

typedef unsigned TraceHandle;
typedef uint8 StopCallbackID;
typedef uint8 GlobalCallbackID;
typedef struct StopMsg {
   uint8 msg[8];
} StopMsg;

#if defined(MONITOR) || defined (VMKERNEL)

/*
 * Include stdarg.h so that we can define sprintf and vsprintf to actually
 * match the implementation.  This also matches the linux kernel header file
 * definition making life easier for the vmklinux module.
 */
#include <stdarg.h>

extern int sprintf(char * buf, const char * fmt, ...);
extern int vsprintf(char *buf, const char *, va_list);
extern int snprintf(char *str, size_t size, const char *format, ...);
extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);

#endif /* MONITOR */

/*
 * Standardize some posix names on Windows.
 */
#ifdef _WIN32

#define snprintf _snprintf
#define	vsnprintf	_vsnprintf

#define strcasecmp(_s1,_s2)   _stricmp((_s1),(_s2))
#define strncasecmp(_s1,_s2,_n)   _strnicmp((_s1),(_s2),(_n))
#define sleep(sec)   Sleep((sec) * 1000)
#define usleep(usec) Sleep(CEILING(usec, 1000))

typedef int pid_t;
#define	getpid		_getpid

#define open _open

#define	access        _access
#define       F_OK          0
#define       X_OK          1
#define       W_OK          2
#define       R_OK          4

#define PATH_MAX 256

#endif /* _WIN32 */

#endif //_VM_TYPES_H_
